package org.jboss.cache.eviction;

import junit.framework.TestCase;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jboss.cache.AbstractTreeCacheListener;
import org.jboss.cache.CacheException;
import org.jboss.cache.ExtendedTreeCacheListener;
import org.jboss.cache.Fqn;
import org.jboss.cache.PropertyConfigurator;
import org.jboss.cache.TreeCache;

import EDU.oswego.cs.dl.util.concurrent.BoundedLinkedQueue;
import EDU.oswego.cs.dl.util.concurrent.Callable;
import EDU.oswego.cs.dl.util.concurrent.Executor;
import EDU.oswego.cs.dl.util.concurrent.FutureResult;
import EDU.oswego.cs.dl.util.concurrent.ThreadedExecutor;

public class OptimisticReplicatedEvictionTest extends TestCase
{
   private static final Log log = LogFactory.getLog(OptimisticReplicatedEvictionTest.class);

    private TreeCache cache1;
    private TreeCache cache2;
    
    private BoundedLinkedQueue queue = new BoundedLinkedQueue();
    private static final Fqn SHUTDOWN_FQN = new Fqn("time-to-shutdown");
    
    protected void setUp() throws Exception
    {
        super.setUp();
        
        cache1 = new TreeCache();
        PropertyConfigurator config1 = new PropertyConfigurator();
        config1.configure(cache1, "META-INF/optimistic-replicated-eviction.xml");
        cache1.startService();
                
        cache2 = new TreeCache();
        PropertyConfigurator config2 = new PropertyConfigurator();
        config2.configure(cache2, "META-INF/optimistic-replicated-eviction.xml");
        RemovalThread removalThread = new RemovalThread(cache2);
        log.trace("Starting the removal thread...");
        removalThread.start();
        EvictionListener listener = new EvictionListener();
        log.trace("Listening for evictions...");
        cache2.addTreeCacheListener(listener);
        cache2.startService();
    }

    protected void tearDown() throws Exception
    {
        if (cache1 != null)
        {
            cache1.stopService();
            cache1 = null;
        }
        super.tearDown();
    }
    
    public void testClusteredEvictionToClearVersionInfo() throws Exception 
    {
       cache1.put( "/timeToLiveBased/test000", "k-allele", "v1-allele" ); // this could be from any node 
       cache1.put( "/timeToLiveBased/test000", "k-allele", "v2-allele" ); // v2

       Callable c = new Callable() {
         public Object call() throws Exception {
            Thread.sleep(8 * 1000);
            Object ret = cache2.put("/timeToLiveBased/test000", "k-allele", "v3-allele" ); // will always fail during replication to A:
            if (ret instanceof Exception) throw (Exception) ret;
            return null;
         }
      };
      
      Executor executor = new ThreadedExecutor();
      FutureResult f = new FutureResult();
      Runnable command = f.setter(c);
      executor.execute(command);
       
      for (;;) 
      {
         if (f.isReady()) 
         {
            f.get();
            break;
         }
         else 
         {
            Thread.sleep(2 * 1000);
         }
         cache1.get("/timeToLiveBased/test000", "k-allele"); // refresh the eviction timer so the node will never be evicted on A
      }
       
       assertEquals(cache1.get("/timeToLiveBased/test000", "k-allele"), "v3-allele");
       assertEquals(cache2.get("/timeToLiveBased/test000", "k-allele"), "v3-allele");
    }
    
    // Jimmy's cluster wide removal workaround
    /**
     * An inner class that serves as a listener for cache eviction events.
     */
    private class EvictionListener extends AbstractTreeCacheListener implements ExtendedTreeCacheListener
    {
        public void nodeEvict(Fqn fqn, boolean pre)
        {
            // make sure we act *after* eviction takes place
            if (!pre)
            {
                if (log.isTraceEnabled())
                {
                    StringBuilder builder = new StringBuilder("Adding ");
                    builder.append(fqn);
                    builder.append(" for removal...");                    
                    log.trace(builder);                        
                }
                try {
                  queue.put(fqn);
               } catch (InterruptedException e) {
                  Thread.currentThread().interrupt();
               }
            }
        }
    }

    /**
     * An inner class that serves as the thread that calls the cluster-wide
     * remove operation as cache listener event message threads are cache
     * blocking.
     */
    private class RemovalThread extends Thread
    {
       private final TreeCache cache;
       
        public RemovalThread(TreeCache cache)
        {
            setName("Cache Eviction Remover");
            this.cache = cache;
        }
        
        public void run()
        {
            log.trace("The cache eviction remover thread is starting...");
            
            while (! isInterrupted())
            {
                Fqn fqn = null;
                
                try
                {
                    fqn = (Fqn) queue.take();
                    
                    // if we're shutting down, do nothing...
                    if (fqn != SHUTDOWN_FQN)
                    {                    
                        if (log.isTraceEnabled())
                        {
                            log.trace("Removing " + fqn);
                        }
                    
                        cache.remove(fqn);
                    }
                }
                catch (CacheException ce)
                {
                    // if this happens, data will remain in remote caches so
                    // indicate it as WARN or even ERROR
                    StringBuilder builder = new StringBuilder("Failed to remove ");
                    builder.append(fqn);
                    builder.append(" from remote caches:");                    
                    log.warn(builder, ce);
                }
                catch (InterruptedException ie)
                {
                    Thread.currentThread().interrupt();
                    // we were interrupted in queue.take() so stop
                    break;
                }             
            }
            log.trace("The cache eviction remover thread is stopping...");
        }
    }    
}
